# SPDX-License-Identifier: LGPL-2.1-or-later

import json
import tempfile
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import cast

from mkosi.archive import extract_tar
from mkosi.config import Architecture, Config
from mkosi.context import Context
from mkosi.curl import curl
from mkosi.distribution import Distribution, DistributionInstaller, PackageType, join_mirror
from mkosi.installer.apt import Apt, AptRepository
from mkosi.log import die
from mkosi.run import run, workdir
from mkosi.sandbox import umask


class Installer(DistributionInstaller, distribution=Distribution.debian):
    @classmethod
    def pretty_name(cls) -> str:
        return "Debian"

    @classmethod
    def filesystem(cls) -> str:
        return "ext4"

    @classmethod
    def package_type(cls) -> PackageType:
        return PackageType.deb

    @classmethod
    def default_release(cls) -> str:
        return "testing"

    @classmethod
    def package_manager(cls, config: Config) -> type[Apt]:
        return Apt

    @classmethod
    def repositories(cls, context: Context, for_image: bool = False) -> Iterable[AptRepository]:
        types = ("deb", "deb-src")
        components = ("main", *context.config.repositories)
        mirror = None if for_image else context.config.mirror
        snapshot = None if for_image else context.config.snapshot

        if context.config.local_mirror and not for_image:
            yield AptRepository(
                types=("deb",),
                url=context.config.local_mirror,
                suite=context.config.release,
                components=("main",),
                signedby=None,
            )
            return

        if mirror:
            pass
        elif snapshot:
            mirror = "https://snapshot.debian.org"
        else:
            mirror = "http://deb.debian.org"

        if snapshot:
            url = join_mirror(mirror, f"archive/debian/{snapshot}")
        else:
            url = join_mirror(mirror, "debian")

        signedby = Path("/usr/share/keyrings/debian-archive-keyring.gpg")

        yield AptRepository(
            types=types,
            url=url,
            suite=context.config.release,
            components=components,
            signedby=signedby,
        )

        # Debug repos are typically not mirrored.
        if snapshot:
            url = join_mirror(mirror, f"archive/debian-debug/{snapshot}")
        else:
            url = join_mirror(mirror, "debian-debug")

        yield AptRepository(
            types=types,
            url=url,
            suite=f"{context.config.release}-debug",
            components=components,
            signedby=signedby,
        )

        if context.config.release in ("unstable", "sid"):
            return

        if not snapshot:
            yield AptRepository(
                types=types,
                url=join_mirror(mirror, "debian"),
                suite=f"{context.config.release}-updates",
                components=components,
                signedby=signedby,
            )

        # Security updates repos are never mirrored.
        if snapshot:
            url = join_mirror(mirror, f"archive/debian-security/{snapshot}")
        else:
            url = join_mirror(mirror, "debian-security")

        yield AptRepository(
            types=types,
            url=url,
            suite=f"{context.config.release}-security",
            components=components,
            signedby=signedby,
        )

    @classmethod
    def setup(cls, context: Context) -> None:
        Apt.setup(context, list(cls.repositories(context)))

    @classmethod
    def install(cls, context: Context) -> None:
        # Instead of using debootstrap, we replicate its core functionality here. Because dpkg does not have
        # an option to delay running pre-install maintainer scripts when it installs a package, it's
        # impossible to use apt directly to bootstrap a Debian chroot since dpkg will try to run a maintainer
        # script which depends on some basic tool to be available in the chroot from a deb which hasn't been
        # unpacked yet, causing the script to fail. To avoid these issues, we have to extract all the
        # essential debs first, and only then run the maintainer scripts for them.

        # First, we set up merged usr.  This list is taken from
        # https://salsa.debian.org/installer-team/debootstrap/-/blob/master/functions#L1369.
        subdirs = ["bin", "sbin", "lib"] + {
            "amd64"       : ["lib32", "lib64", "libx32"],
            "i386"        : ["lib64", "libx32"],
            "mips"        : ["lib32", "lib64"],
            "mipsel"      : ["lib32", "lib64"],
            "mips64el"    : ["lib32", "lib64", "libo32"],
            "loong64"     : ["lib32", "lib64"],
            "powerpc"     : ["lib64"],
            "ppc64"       : ["lib32", "lib64"],
            "ppc64el"     : ["lib64"],
            "s390x"       : ["lib32"],
            "sparc"       : ["lib64"],
            "sparc64"     : ["lib32", "lib64"],
            "x32"         : ["lib32", "lib64", "libx32"],
        }.get(
            context.config.distribution.installer.architecture(context.config.architecture), []
        )  # fmt: skip

        with umask(~0o755):
            for d in subdirs:
                (context.root / d).symlink_to(f"usr/{d}")
                (context.root / f"usr/{d}").mkdir(parents=True, exist_ok=True)

        # Next, we invoke apt-get install to download all the essential packages. With
        # DPkg::Pre-Install-Pkgs, we specify a shell command that will receive the list of packages that will
        # be installed on stdin.  By configuring Debug::pkgDpkgPm=1, apt-get install will not actually
        # execute any dpkg commands, so all it does is download the essential debs and tell us their full in
        # the apt cache without actually installing them.
        with tempfile.NamedTemporaryFile(mode="r") as f:
            Apt.invoke(
                context,
                "install",
                [
                    "-oDebug::pkgDPkgPm=1",
                    f"-oDPkg::Pre-Install-Pkgs::=cat >{workdir(Path(f.name))}",
                    "?essential",
                    "base-files",
                ],
                options=["--bind", f.name, workdir(Path(f.name))],
            )

            essential = f.read().strip().splitlines()

        # Now, extract the debs to the chroot by first extracting the sources tar file out of the deb and
        # then extracting the tar file into the chroot.

        for deb in essential:
            # If a deb path is in the form of "/var/cache/apt/<deb>", we transform it to the corresponding
            # path in mkosi's package cache directory. If it's relative to /repository, we transform it to
            # the corresponding path in mkosi's local package repository. Otherwise, we use the path as is.
            if Path(deb).is_relative_to("/var/cache"):
                path = context.config.package_cache_dir_or_default() / Path(deb).relative_to("/var")
            elif Path(deb).is_relative_to("/repository"):
                path = context.repository / Path(deb).relative_to("/repository")
            else:
                path = Path(deb)

            with open(path, "rb") as i, tempfile.NamedTemporaryFile() as o:
                run(
                    ["dpkg-deb", "--fsys-tarfile", "/dev/stdin"],
                    stdin=i,
                    stdout=o,
                    sandbox=context.sandbox(),
                )
                extract_tar(
                    Path(o.name),
                    context.root,
                    log=False,
                    options=(
                        [f"--exclude=./{glob}" for glob in Apt.documentation_exclude_globs]
                        if not context.config.with_docs
                        else []
                    ),
                    sandbox=context.sandbox,
                )

        # Finally, run apt to properly install packages in the chroot without having to worry that maintainer
        # scripts won't find basic tools that they depend on.

        cls.install_packages(
            context, [Path(deb).name.partition("_")[0].removesuffix(".deb") for deb in essential]
        )

        fixup_os_release(context)

    @classmethod
    def install_packages(
        cls,
        context: Context,
        packages: Sequence[str],
        *,
        apivfs: bool = True,
        allow_downgrade: bool = False,
    ) -> None:
        super().install_packages(context, packages, apivfs=apivfs, allow_downgrade=allow_downgrade)

        if "apt" in packages:
            install_apt_sources(context, cls.repositories(context, for_image=True))

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.arm64:       "arm64",
            Architecture.arm:         "armhf",
            Architecture.alpha:       "alpha",
            Architecture.x86_64:      "amd64",
            Architecture.x86:         "i386",
            Architecture.ia64:        "ia64",
            Architecture.loongarch64: "loong64",
            Architecture.mips64_le:   "mips64el",
            Architecture.mips_le:     "mipsel",
            Architecture.parisc:      "hppa",
            Architecture.ppc64_le:    "ppc64el",
            Architecture.ppc64:       "ppc64",
            Architecture.riscv64:     "riscv64",
            Architecture.s390x:       "s390x",
            Architecture.s390:        "s390",
        }.get(arch)  # fmt: skip

        if not a:
            die(f"Architecture {arch} is not supported by {cls.pretty_name()}")

        return a

    @classmethod
    def latest_snapshot(cls, config: Config) -> str:
        url = join_mirror(config.mirror or "https://snapshot.debian.org", "mr/timestamp")
        return cast(str, json.loads(curl(config, url))["result"]["debian"][-1])

    @classmethod
    def is_kernel_package(cls, package: str) -> bool:
        return package.startswith("linux-image-")


def install_apt_sources(context: Context, repos: Iterable[AptRepository]) -> None:
    sources = context.root / f"etc/apt/sources.list.d/{context.config.release}.sources"
    if not sources.exists():
        with umask(~0o755):
            sources.parent.mkdir(parents=True, exist_ok=True)
        with umask(~0o644), sources.open("w") as f:
            for repo in repos:
                f.write(str(repo))


def fixup_os_release(context: Context) -> None:
    if context.config.release not in ("unstable", "sid"):
        return

    # Debian being Debian means we need to special case handling os-release. Fix the content to actually
    # match what we are building, and set up a diversion so that dpkg doesn't overwrite it on package
    # updates.  Upstream bug report: https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=1008735.
    for candidate in ["etc/os-release", "usr/lib/os-release", "usr/lib/initrd-release"]:
        osrelease = context.root / candidate
        newosrelease = osrelease.with_suffix(".new")

        if not osrelease.is_file():
            continue

        if osrelease.is_symlink() and candidate != "etc/os-release":
            continue

        with osrelease.open("r") as old, newosrelease.open("w") as new:
            for line in old.readlines():
                if line.startswith("VERSION_CODENAME="):
                    new.write("VERSION_CODENAME=sid\n")
                else:
                    new.write(line)

        # On dpkg distributions we cannot simply overwrite /etc/os-release as it is owned by a package.  We
        # need to set up a diversion first, so that it is not overwritten by package updates.  We do this for
        # /etc/os-release as that will be overwritten on package updates and has precedence over
        # /usr/lib/os-release, and ignore the latter and assume that if an usr-only image is built then the
        # package manager will not run on it.
        if candidate == "etc/os-release":
            run(
                [
                    "dpkg-divert",
                    "--quiet",
                    "--root=/buildroot",
                    "--local",
                    "--add",
                    "--rename",
                    "--divert",
                    f"/{candidate}.dpkg",
                    f"/{candidate}",
                ],
                sandbox=context.sandbox(options=context.rootoptions()),
            )

        newosrelease.rename(osrelease)



# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import sys

from mkosi.config import Args, Config, OutputFormat
from mkosi.log import complete_step, die
from mkosi.run import run
from mkosi.user import become_root_cmd


def run_burn(args: Args, config: Config) -> None:
    if config.output_format not in (OutputFormat.disk, OutputFormat.esp):
        die(f"{config.output_format} images cannot be burned to disk")

    if not args.cmdline:
        die("Please specify a device to burn the image to", hint="For example /dev/disk/by-id/usb-foobar")

    fname = config.output_dir_or_cwd() / config.output

    if len(args.cmdline) != 1:
        die("Expected device argument.")

    cmd = [
        "systemd-repart",
        "--no-pager",
        "--pretty=no",
        "--offline=yes",
        "--empty=force",
        "--dry-run=no",
        "--definitions=/",
        f"--copy-from={fname}",
        *args.cmdline,
    ]

    with complete_step("Burning 🔥🔥🔥 to medium…", "Burnt. 🔥🔥🔥"):
        run(
            cmd,
            stdin=sys.stdin,
            stdout=sys.stdout,
            env=os.environ | config.finalize_environment(),
            log=False,
            sandbox=config.sandbox(
                devices=True,
                network=True,
                relaxed=True,
                options=["--same-dir"],
            ),
            setup=become_root_cmd(),
        )



# SPDX-License-Identifier: LGPL-2.1-or-later

import datetime
import tempfile
from collections.abc import Iterable
from pathlib import Path

from mkosi.archive import extract_tar
from mkosi.config import Architecture, Config
from mkosi.context import Context
from mkosi.curl import curl
from mkosi.distribution import Distribution, DistributionInstaller, PackageType, join_mirror
from mkosi.installer.pacman import Pacman, PacmanRepository
from mkosi.log import complete_step, die


class Installer(DistributionInstaller, distribution=Distribution.arch):
    @classmethod
    def pretty_name(cls) -> str:
        return "Arch Linux"

    @classmethod
    def filesystem(cls) -> str:
        return "ext4"

    @classmethod
    def package_type(cls) -> PackageType:
        return PackageType.pkg

    @classmethod
    def default_release(cls) -> str:
        return "rolling"

    @classmethod
    def package_manager(cls, config: "Config") -> type[Pacman]:
        return Pacman

    @classmethod
    def keyring(cls, context: Context) -> None:
        if context.config.repository_key_fetch:
            with (
                complete_step(f"Downloading {cls.pretty_name()} keyring"),
                tempfile.TemporaryDirectory() as d,
            ):
                curl(
                    context.config,
                    "https://archlinux.org/packages/core/any/archlinux-keyring/download",
                    output_dir=Path(d),
                )
                extract_tar(
                    next(Path(d).iterdir()),
                    context.sandbox_tree,
                    dirs=["usr/share/pacman/keyrings"],
                    sandbox=context.sandbox,
                )

        Pacman.keyring(context)

    @classmethod
    def setup(cls, context: Context) -> None:
        Pacman.setup(context, list(cls.repositories(context)))

    @classmethod
    def install(cls, context: Context) -> None:
        cls.install_packages(context, ["filesystem"], apivfs=False)

    @classmethod
    def repositories(cls, context: Context) -> Iterable[PacmanRepository]:
        if context.config.local_mirror:
            yield PacmanRepository("core", context.config.local_mirror)
        else:
            if context.config.architecture.is_arm_variant():
                if context.config.snapshot and not context.config.mirror:
                    die("There is no known public mirror for snapshots of Arch Linux ARM")

                mirror = context.config.mirror or "http://mirror.archlinuxarm.org"
            else:
                if context.config.mirror:
                    mirror = context.config.mirror
                elif context.config.snapshot:
                    mirror = "https://archive.archlinux.org"
                else:
                    mirror = "https://fastly.mirror.pkgbuild.com"

            if context.config.snapshot:
                url = join_mirror(mirror, f"repos/{context.config.snapshot}/$repo/os/$arch")
            elif context.config.architecture.is_arm_variant():
                url = join_mirror(mirror, "$arch/$repo")
            else:
                url = join_mirror(mirror, "$repo/os/$arch")

            # Testing repositories have to go before regular ones to to take precedence.
            repos = [
                repo
                for repo in (
                    "core-testing",
                    "core-testing-debug",
                    "extra-testing",
                    "extra-testing-debug",
                    "core-debug",
                    "extra-debug",
                    "multilib-testing",
                    "multilib",
                )
                if repo in context.config.repositories
            ] + ["core", "extra"]

            if context.config.architecture.is_arm_variant():
                repos += ["alarm"]

            for repo in repos:
                yield PacmanRepository(repo, url)

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.x86_64: "x86_64",
            Architecture.arm64:  "aarch64",
            Architecture.arm:    "armv7h",
        }.get(arch)  # fmt: skip

        if not a:
            die(f"Architecture {a} is not supported by Arch Linux")

        return a

    @classmethod
    def latest_snapshot(cls, config: Config) -> str:
        url = join_mirror(config.mirror or "https://archive.archlinux.org", "repos/last/lastsync")
        return datetime.datetime.fromtimestamp(int(curl(config, url)), datetime.timezone.utc).strftime(
            "%Y/%m/%d"
        )

    @classmethod
    def is_kernel_package(cls, package: str) -> bool:
        return package in ("linux", "linux-lts", "linux-zen", "linux-hardened", "linux-rt", "linux-rt-lts")



# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import os
import subprocess
import sys
import tempfile
from pathlib import Path

from mkosi.config import Args, ArtifactOutput, Config, OutputFormat
from mkosi.log import die
from mkosi.run import run, workdir
from mkosi.user import become_root_cmd
from mkosi.util import PathString, flatten


def run_sysupdate(args: Args, config: Config) -> None:
    if not config.sysupdate_dir:
        die(
            "No sysupdate definitions directory specified",
            hint="Specify a directory containing systemd-sysupdate transfer definitions with "
            "SysupdateDirectory=",
        )

    if not (sysupdate := config.find_binary("systemd-sysupdate", "/usr/lib/systemd/systemd-sysupdate")):
        die("Could not find systemd-sysupdate")

    with contextlib.ExitStack() as stack:
        if config.tools() != Path("/"):
            # We explicitly run this without a sandbox, because / has to be the original root mountpoint for
            # bootctl --print-root-device to work properly.
            blockdev = run(["bootctl", "--print-root-device"], stdout=subprocess.PIPE).stdout.strip()

            tmp = stack.enter_context(tempfile.TemporaryDirectory())
            # If /run/systemd/volatile-root exists, systemd skips its root block device detection logic and
            # uses whatever block device /run/systemd/volatile-root points to instead. Let's make use of that
            # when using a tools tree as in that case the block device detection logic doesn't work properly.
            (Path(tmp) / "volatile-root").symlink_to(blockdev)
        else:
            tmp = None

        if (
            config.output_format == OutputFormat.disk
            and ArtifactOutput.partitions not in config.split_artifacts
        ):
            old = {p for p in config.output_dir_or_cwd().iterdir() if p.is_file()}

            # If we didn't generate split partitions as part of the image build, let's do it now.
            run(
                [
                    "systemd-repart",
                    "--no-pager",
                    "--pretty=no",
                    "--split=yes",
                    *([f"--definitions={workdir(d)}" for d in config.repart_dirs]),
                    workdir(config.output_dir_or_cwd() / config.output_with_format),
                ],
                sandbox=config.sandbox(
                    options=[
                        "--bind", config.output_dir_or_cwd(), workdir(config.output_dir_or_cwd()),
                        *flatten(["--ro-bind", os.fspath(d), workdir(d)] for d in config.repart_dirs),
                    ],
                ),
            )  # fmt: skip

            for p in config.output_dir_or_cwd().iterdir():
                if p not in old and p.is_file():
                    stack.callback(p.unlink)

        cmd: list[PathString] = [
            sysupdate,
            "--definitions", config.sysupdate_dir,
            "--transfer-source", config.output_dir_or_cwd(),
            *args.cmdline,
        ]  # fmt: skip

        run(
            cmd,
            stdin=sys.stdin,
            stdout=sys.stdout,
            env=os.environ | config.finalize_environment(),
            log=False,
            sandbox=config.sandbox(
                devices=True,
                network=True,
                relaxed=True,
                options=[
                    *(["--bind", "/boot", "/boot"] if Path("/boot").exists() else []),
                    *(["--bind", "/efi", "/efi"] if Path("/efi").exists() else []),
                    *(
                        [
                            # Make sure systemd-sysupdate parses os-release from the host and not the tools
                            # tree.
                            "--bind", "/usr/lib/os-release", "/usr/lib/os-release",
                            "--bind", tmp, "/run/systemd",
                        ]
                        if tmp
                        else []
                    ),
                    "--same-dir",
                ],
            ),
            setup=become_root_cmd(),
        )  # fmt: skip



# SPDX-License-Identifier: LGPL-2.1-or-later

import textwrap
from collections.abc import Sequence
from pathlib import Path
from typing import Optional

from mkosi.config import Cacheonly, Config
from mkosi.context import Context
from mkosi.installer import PackageManager
from mkosi.installer.rpm import RpmRepository, rpm_cmd
from mkosi.log import ARG_DEBUG
from mkosi.run import CompletedProcess, run, workdir
from mkosi.util import _FILE, PathString


class Dnf(PackageManager):
    @classmethod
    def executable(cls, config: Config) -> str:
        # Allow the user to override autodetection with an environment variable
        dnf = config.finalize_environment().get("MKOSI_DNF")
        return Path(dnf or config.find_binary("dnf5") or "dnf").name

    @classmethod
    def subdir(cls, config: Config) -> Path:
        return Path("libdnf5" if cls.executable(config) == "dnf5" else "dnf")

    @classmethod
    def package_subdirs(cls, cache: Path) -> list[tuple[Path, Path]]:
        dirs = [p for p in cache.iterdir() if p.is_dir() and "-" in p.name and "mkosi" not in p.name]
        return [
            (
                # If the package cache directory is set to /var, we need to make sure we look up packages
                # where they were stored by dnf, so don't do any special handling in that case.
                d.relative_to(cache) / "packages"
                if cache.parent == Path("/var/cache")
                # Cache directories look like <repo-id>-<baseurl-hash> so let's strip off the hash to reuse
                # the same package cache directory regardless of baseurl.
                else Path("packages") / d.name[: d.name.rfind("-")],
                d.relative_to(cache) / "packages",
            )
            for d in dirs
        ]

    @classmethod
    def package_globs(cls) -> list[str]:
        return ["*.rpm"]

    @classmethod
    def scripts(cls, context: Context) -> dict[str, list[PathString]]:
        return {
            "dnf": cls.apivfs_script_cmd(context) + cls.env_cmd(context) + cls.cmd(context),
            "rpm": cls.apivfs_script_cmd(context) + rpm_cmd(),
            "mkosi-install":   ["dnf", "install"],
            "mkosi-upgrade":   ["dnf", "upgrade"],
            "mkosi-remove":    ["dnf", "remove"],
            "mkosi-reinstall": ["dnf", "reinstall"],
        }  # fmt: skip

    @classmethod
    def setup(
        cls,
        context: Context,
        repositories: Sequence[RpmRepository],
        filelists: bool = True,
        metadata_expire: Optional[str] = None,
    ) -> None:
        (context.sandbox_tree / "etc/dnf/vars").mkdir(parents=True, exist_ok=True)
        (context.sandbox_tree / "etc/yum.repos.d").mkdir(parents=True, exist_ok=True)

        config = context.sandbox_tree / "etc/dnf/dnf.conf"

        if not config.exists():
            config.parent.mkdir(exist_ok=True, parents=True)
            with config.open("w") as f:
                # Make sure we download filelists so all dependencies can be resolved.
                # See https://bugzilla.redhat.com/show_bug.cgi?id=2180842
                if cls.executable(context.config) == "dnf5" and filelists:
                    f.write("[main]\noptional_metadata_types=filelists\n")

        # The CentOS Hyperscale ships a COW plugin for dnf that's disabled by default. Let's enable it so we
        # can take advantage of faster rpm package installations.
        reflink = context.sandbox_tree / "etc/dnf/plugins/reflink.conf"
        if not reflink.exists():
            reflink.parent.mkdir(parents=True, exist_ok=True)
            reflink.write_text(
                textwrap.dedent(
                    """\
                    [main]
                    enabled=1
                    """
                )
            )

        # The versionlock plugin will fail if enabled without a configuration file so lets' write a noop
        # configuration file to make it happy which can be overridden by users.
        versionlock = context.sandbox_tree / "etc/dnf/plugins/versionlock.conf"
        if not versionlock.exists():
            versionlock.parent.mkdir(parents=True, exist_ok=True)
            versionlock.write_text(
                textwrap.dedent(
                    """\
                    [main]
                    enabled=0
                    locklist=/dev/null
                    """
                )
            )

        repofile = context.sandbox_tree / "etc/yum.repos.d/mkosi.repo"
        if not repofile.exists():
            repofile.parent.mkdir(exist_ok=True, parents=True)
            with repofile.open("w") as f:
                for repo in repositories:
                    f.write(
                        textwrap.dedent(
                            f"""\
                            [{repo.id}]
                            name={repo.id}
                            {repo.url}
                            gpgcheck=1
                            enabled={int(repo.enabled)}
                            """
                        )
                    )

                    if repo.sslcacert:
                        f.write(f"sslcacert={repo.sslcacert}\n")
                    if repo.sslclientcert:
                        f.write(f"sslclientcert={repo.sslclientcert}\n")
                    if repo.sslclientkey:
                        f.write(f"sslclientkey={repo.sslclientkey}\n")
                    if repo.priority:
                        f.write(f"priority={repo.priority}\n")
                    if metadata_expire:
                        f.write(f"metadata_expire={metadata_expire}\n")

                    for i, url in enumerate(repo.gpgurls):
                        f.write("gpgkey=" if i == 0 else len("gpgkey=") * " ")
                        f.write(f"{url}\n")

                    f.write("\n")

    @classmethod
    def finalize_environment(cls, context: Context) -> dict[str, str]:
        return super().finalize_environment(context) | {
            "RPM_FORCE_DEBIAN": "1",
        }

    @classmethod
    def cmd(
        cls,
        context: Context,
        cached_metadata: bool = True,
    ) -> list[PathString]:
        dnf = cls.executable(context.config)

        cmdline: list[PathString] = [
            dnf,
            "--assumeyes",
            "--best",
            f"--releasever={context.config.release}",
            "--installroot=/buildroot",
            "--setopt=keepcache=1",
            "--setopt=logdir=/var/log",
            f"--setopt=cachedir=/var/cache/{cls.subdir(context.config)}",
            f"--setopt=install_weak_deps={int(context.config.with_recommends)}",
            "--setopt=check_config_file_age=0",
            "--setopt=persistdir=/buildroot/var/lib/dnf",
        ]

        if ARG_DEBUG.get():
            cmdline += ["--setopt=debuglevel=10"]

        if not context.config.repository_key_check:
            cmdline += ["--nogpgcheck"]

        if context.config.repositories:
            opt = "--enable-repo" if dnf == "dnf5" else "--enablerepo"
            cmdline += [f"{opt}={repo}" for repo in context.config.repositories]

        if context.config.cacheonly == Cacheonly.always:
            cmdline += ["--cacheonly"]
        elif cached_metadata:
            cmdline += ["--setopt=metadata_expire=never"]
            if dnf == "dnf5":
                cmdline += ["--setopt=cacheonly=metadata"]

        if not context.config.architecture.is_native():
            cmdline += [f"--forcearch={cls.architecture(context)}"]

        if not context.config.with_docs:
            cmdline += ["--no-docs" if dnf == "dnf5" else "--nodocs"]

        if dnf == "dnf5":
            cmdline += ["--use-host-config"]
        else:
            cmdline += [
                "--config=/etc/dnf/dnf.conf",
                "--setopt=reposdir=/etc/yum.repos.d",
                "--setopt=varsdir=/etc/dnf/vars",
            ]

        if context.config.proxy_url:
            cmdline += [f"--setopt=proxy={context.config.proxy_url}"]
        if context.config.proxy_peer_certificate:
            cmdline += ["--setopt=proxy_sslcacert=/proxy.cacert"]
        if context.config.proxy_client_certificate:
            cmdline += ["--setopt=proxy_sslclientcert=/proxy.clientcert"]
        if context.config.proxy_client_key:
            cmdline += ["--setopt=proxy_sslclientkey=/proxy.clientkey"]

        return cmdline

    @classmethod
    def invoke(
        cls,
        context: Context,
        operation: str,
        arguments: Sequence[str] = (),
        *,
        apivfs: bool = False,
        stdout: _FILE = None,
        cached_metadata: bool = True,
    ) -> CompletedProcess:
        try:
            return run(
                cls.cmd(context, cached_metadata=cached_metadata) + [operation, *arguments],
                sandbox=cls.sandbox(context, apivfs=apivfs),
                env=cls.finalize_environment(context),
                stdout=stdout,
            )
        finally:
            # dnf interprets the log directory relative to the install root so there's nothing we can do but
            # to remove the log files from the install root afterwards.
            if (context.root / "var/log").exists():
                for p in (context.root / "var/log").iterdir():
                    if any(p.name.startswith(prefix) for prefix in ("dnf", "hawkey", "yum")):
                        p.unlink()

    @classmethod
    def install(
        cls,
        context: Context,
        packages: Sequence[str],
        *,
        apivfs: bool = True,
        allow_downgrade: bool = False,
    ) -> None:
        arguments = []

        if allow_downgrade and Dnf.executable(context.config) == "dnf5":
            arguments += ["--allow-downgrade"]

        arguments += [*packages]

        cls.invoke(context, "install", arguments, apivfs=apivfs)

    @classmethod
    def remove(cls, context: Context, packages: Sequence[str]) -> None:
        cls.invoke(context, "remove", packages, apivfs=True)

    @classmethod
    def sync(cls, context: Context, force: bool, arguments: Sequence[str] = ()) -> None:
        cls.invoke(
            context,
            "makecache",
            arguments=[*(["--refresh"] if force else []), *arguments],
            cached_metadata=False,
        )

    @classmethod
    def createrepo(cls, context: Context) -> None:
        run(
            ["createrepo_c", workdir(context.repository)],
            sandbox=context.sandbox(options=["--bind", context.repository, workdir(context.repository)]),
        )

        (context.sandbox_tree / "etc/yum.repos.d/mkosi-local.repo").write_text(
            textwrap.dedent(
                """\
                [mkosi]
                name=mkosi
                baseurl=file:///repository
                gpgcheck=0
                metadata_expire=never
                priority=10
                """
            )
        )

        cls.sync(context, force=True, arguments=["--disablerepo=*", "--enablerepo=mkosi"])



# SPDX-License-Identifier: LGPL-2.1-or-later

import os
from collections.abc import Sequence
from contextlib import AbstractContextManager
from pathlib import Path
from typing import Optional

from mkosi.config import Args, Config
from mkosi.util import PathString, flatten


class Context:
    """State related properties."""

    def __init__(
        self,
        args: Args,
        config: Config,
        *,
        workspace: Path,
        resources: Path,
        keyring_dir: Path,
        metadata_dir: Path,
        package_dir: Optional[Path] = None,
    ) -> None:
        self.args = args
        self.config = config
        self.workspace = workspace
        self.resources = resources
        self.keyring_dir = keyring_dir
        self.metadata_dir = metadata_dir
        self.package_dir = package_dir or (self.workspace / "packages")
        self.lowerdirs: list[PathString] = []
        self.upperdir: Optional[PathString] = None
        self.workdir: Optional[PathString] = None

        self.package_dir.mkdir(exist_ok=True)
        self.staging.mkdir()
        self.sandbox_tree.mkdir()
        self.repository.mkdir()
        self.artifacts.mkdir()
        self.install_dir.mkdir()

    @property
    def root(self) -> Path:
        return self.workspace / "root"

    def rootoptions(self, dst: PathString = "/buildroot", *, readonly: bool = False) -> list[str]:
        if self.lowerdirs or self.upperdir:
            return [
                "--overlay-lowerdir", os.fspath(self.root),
                *flatten(["--overlay-lowerdir", os.fspath(lowerdir)] for lowerdir in self.lowerdirs),
                *(
                    ["--overlay-lowerdir" if readonly else "--overlay-upperdir", os.fspath(self.upperdir)]
                    if self.upperdir
                    else []
                ),
                *(["--overlay-workdir", os.fspath(self.workdir)] if self.workdir and not readonly else []),
                "--overlay", os.fspath(dst),
            ]  # fmt: skip
        else:
            return ["--ro-bind" if readonly else "--bind", os.fspath(self.root), os.fspath(dst)]

    @property
    def staging(self) -> Path:
        return self.workspace / "staging"

    @property
    def sandbox_tree(self) -> Path:
        return self.workspace / "sandbox"

    @property
    def repository(self) -> Path:
        return self.workspace / "repository"

    @property
    def artifacts(self) -> Path:
        return self.workspace / "artifacts"

    @property
    def install_dir(self) -> Path:
        return self.workspace / "dest"

    def sandbox(
        self,
        *,
        network: bool = False,
        devices: bool = False,
        scripts: Optional[Path] = None,
        options: Sequence[PathString] = (),
    ) -> AbstractContextManager[list[PathString]]:
        return self.config.sandbox(
            network=network,
            devices=devices,
            scripts=scripts,
            overlay=self.sandbox_tree,
            options=options,
        )



# SPDX-License-Identifier: LGPL-2.1-or-later

import re
import subprocess
import tempfile
from collections.abc import Iterable
from pathlib import Path

from mkosi.config import Architecture, Config
from mkosi.context import Context
from mkosi.curl import curl
from mkosi.distribution import (
    Distribution,
    DistributionInstaller,
    PackageType,
    join_mirror,
)
from mkosi.installer.dnf import Dnf
from mkosi.installer.rpm import RpmRepository, find_rpm_gpgkey, setup_rpm
from mkosi.log import die
from mkosi.util import tuplify

DISTRIBUTION_GPG_KEYS_UPSTREAM = (
    "https://raw.githubusercontent.com/rpm-software-management/distribution-gpg-keys/main/keys/fedora"
)


def read_remote_rawhide_key_symlink(context: Context) -> str:
    # https://fedoraproject.org/fedora.gpg is always outdated when the rawhide key changes. Instead,
    # let's fetch it from distribution-gpg-keys on github if necessary, which is generally up-to-date.
    with tempfile.TemporaryDirectory() as d:
        # The rawhide key is a symlink and github doesn't redirect those to the actual file for some reason
        curl(
            context.config,
            f"{DISTRIBUTION_GPG_KEYS_UPSTREAM}/RPM-GPG-KEY-fedora-rawhide-primary",
            output_dir=Path(d),
        )
        return (Path(d) / "RPM-GPG-KEY-fedora-rawhide-primary").read_text()


@tuplify
def find_fedora_rpm_gpgkeys(context: Context) -> Iterable[str]:
    versionre = re.compile(r"RPM-GPG-KEY-fedora-(\d+)-primary", re.ASCII)
    # ELN uses the rawhide GPG keys.
    release = "rawhide" if context.config.release == "eln" else context.config.release

    if release == "rawhide" and context.config.repository_key_fetch:
        # Rawhide is a moving target and signed with a different GPG key every time a new Fedora release is
        # done. In distribution-gpg-keys this is modeled by a symlink that is continuously updated to point
        # to the current GPG key for rawhide. Of course, this symlink gets outdated when using a locally
        # installed distribution-gpg-keys package. If we're allowed to look up GPG keys remotely, look up the
        # current rawhide version remotely and use the associated remote key.
        key = read_remote_rawhide_key_symlink(context)
        if not (rawhide_will_be := versionre.match(key)):
            die(f"Missing Fedora version in remote rawhide key {key} from distribution-gpg-keys")

        version = int(rawhide_will_be.group(1))
        yield f"{DISTRIBUTION_GPG_KEYS_UPSTREAM}/RPM-GPG-KEY-fedora-{version}-primary"

        # Also use the N+1 key if it exists to avoid issues when rawhide has been moved to the next key but
        # the rawhide symlink in distribution-gpg-keys hasn't been updated yet.
        try:
            with tempfile.TemporaryDirectory() as d:
                curl(
                    context.config,
                    f"{DISTRIBUTION_GPG_KEYS_UPSTREAM}/RPM-GPG-KEY-fedora-{version + 1}-primary",
                    output_dir=Path(d),
                    log=False,
                )

            yield f"{DISTRIBUTION_GPG_KEYS_UPSTREAM}/RPM-GPG-KEY-fedora-{version + 1}-primary"
        except subprocess.CalledProcessError:
            pass

        return

    key = find_rpm_gpgkey(
        context,
        key=f"RPM-GPG-KEY-fedora-{release}-primary",
        fallback=f"{DISTRIBUTION_GPG_KEYS_UPSTREAM}/RPM-GPG-KEY-fedora-{release}-primary",
    )

    yield key

    if release == "rawhide" and (rawhide_will_be := versionre.match(Path(key).name)):
        # When querying the rawhide version remotely, we add the N+1 key as the symlink might not have been
        # updated yet. We do expect the symlink update to happen in reasonable time so we only add the N+1
        # key. When using a locally installed distribution-gpg-keys package on older Fedora versions, there's
        # a non-zero chance that rawhide might already be using the N+2 key. So let's play it safe and add
        # all newer keys in this case.
        version = int(rawhide_will_be.group(1))

        i = 1
        while newerkey := find_rpm_gpgkey(
            context,
            key=f"RPM-GPG-KEY-fedora-{version + i}-primary",
            required=False,
        ):
            yield newerkey
            i += 1


class Installer(DistributionInstaller, distribution=Distribution.fedora):
    @classmethod
    def pretty_name(cls) -> str:
        return "Fedora Linux"

    @classmethod
    def filesystem(cls) -> str:
        return "btrfs"

    @classmethod
    def package_type(cls) -> PackageType:
        return PackageType.rpm

    @classmethod
    def default_release(cls) -> str:
        return "rawhide"

    @classmethod
    def grub_prefix(cls) -> str:
        return "grub2"

    @classmethod
    def package_manager(cls, config: Config) -> type[Dnf]:
        return Dnf

    @classmethod
    def setup(cls, context: Context) -> None:
        setup_rpm(context)
        Dnf.setup(
            context,
            list(cls.repositories(context)),
            filelists=False,
            metadata_expire="6h" if context.config.release in ("eln", "rawhide") else None,
        )

    @classmethod
    def install(cls, context: Context) -> None:
        cls.install_packages(context, ["basesystem"], apivfs=False)

    @classmethod
    def repositories(cls, context: Context) -> Iterable[RpmRepository]:
        gpgurls = find_fedora_rpm_gpgkeys(context)

        if context.config.snapshot and context.config.release != "rawhide":
            die(f"Snapshot= is only supported for rawhide on {cls.pretty_name()}")

        mirror = context.config.mirror
        if not mirror and context.config.snapshot:
            mirror = "https://kojipkgs.fedoraproject.org"

        if context.config.snapshot and mirror != "https://kojipkgs.fedoraproject.org":
            die(
                f"Snapshot= is only supported for {cls.pretty_name()} if Mirror=https://kojipkgs.fedoraproject.org"
            )

        if mirror == "https://kojipkgs.fedoraproject.org" and not context.config.snapshot:
            die(
                f"Snapshot= must be used on {cls.pretty_name()} if Mirror=https://kojipkgs.fedoraproject.org"
            )

        if context.config.local_mirror:
            yield RpmRepository("fedora", f"baseurl={context.config.local_mirror}", gpgurls)
            return

        if context.config.release == "eln":
            mirror = context.config.mirror or "https://dl.fedoraproject.org/pub/eln/1/"

            for repo in ("AppStream", "BaseOS", "Extras", "CRB"):
                url = f"baseurl={join_mirror(mirror, repo)}"
                yield RpmRepository(repo.lower(), f"{url}/$basearch/os", gpgurls)
                yield RpmRepository(
                    f"{repo.lower()}-debuginfo", f"{url}/$basearch/debug/tree", gpgurls, enabled=False
                )
                yield RpmRepository(f"{repo.lower()}-source", f"{url}/source/tree", gpgurls, enabled=False)
        elif mirror:
            if mirror == "https://kojipkgs.fedoraproject.org":
                subdir = f"compose/{context.config.release}"
            else:
                subdir = "linux/"
                subdir += "development" if context.config.release == "rawhide" else "releases"
                subdir += "/$releasever"

            if context.config.snapshot:
                subdir += f"/Fedora-{context.config.release.capitalize()}-{context.config.snapshot}/compose"

            url = f"baseurl={join_mirror(mirror, f'{subdir}/Everything')}"
            yield RpmRepository("fedora", f"{url}/$basearch/os", gpgurls)
            yield RpmRepository("fedora-debuginfo", f"{url}/$basearch/debug/tree", gpgurls, enabled=False)
            yield RpmRepository("fedora-source", f"{url}/source/tree", gpgurls, enabled=False)

            if context.config.release != "rawhide":
                url = f"baseurl={join_mirror(mirror, 'linux/updates/$releasever/Everything')}"
                yield RpmRepository("updates", f"{url}/$basearch", gpgurls)
                yield RpmRepository("updates-debuginfo", f"{url}/$basearch/debug", gpgurls, enabled=False)
                yield RpmRepository("updates-source", f"{url}/source/tree", gpgurls, enabled=False)

                url = f"baseurl={join_mirror(mirror, 'linux/updates/testing/$releasever/Everything')}"
                yield RpmRepository("updates-testing", f"{url}/$basearch", gpgurls, enabled=False)
                yield RpmRepository(
                    "updates-testing-debuginfo", f"{url}/$basearch/debug", gpgurls, enabled=False
                )
                yield RpmRepository("updates-testing-source", f"{url}/source/tree", gpgurls, enabled=False)
        else:
            url = "metalink=https://mirrors.fedoraproject.org/metalink?arch=$basearch"
            yield RpmRepository("fedora", f"{url}&repo=fedora-$releasever", gpgurls)
            yield RpmRepository(
                "fedora-debuginfo", f"{url}&repo=fedora-debug-$releasever", gpgurls, enabled=False
            )
            yield RpmRepository(
                "fedora-source", f"{url}&repo=fedora-source-$releasever", gpgurls, enabled=False
            )

            if context.config.release != "rawhide":
                yield RpmRepository("updates", f"{url}&repo=updates-released-f$releasever", gpgurls)
                yield RpmRepository(
                    "updates-debuginfo",
                    f"{url}&repo=updates-released-debug-f$releasever",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    "updates-source",
                    f"{url}&repo=updates-released-source-f$releasever",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    "updates-testing", f"{url}&repo=updates-testing-f$releasever", gpgurls, enabled=False
                )
                yield RpmRepository(
                    "updates-testing-debuginfo",
                    f"{url}&repo=updates-testing-debug-f$releasever",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    "updates-testing-source",
                    f"{url}&repo=updates-testing-source-f$releasever",
                    gpgurls,
                    enabled=False,
                )

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.arm64:     "aarch64",
            Architecture.mips64_le: "mips64el",
            Architecture.mips_le:   "mipsel",
            Architecture.ppc64_le:  "ppc64le",
            Architecture.riscv64:   "riscv64",
            Architecture.s390x:     "s390x",
            Architecture.x86_64:    "x86_64",
        }.get(arch)  # fmt: skip

        if not a:
            die(f"Architecture {a} is not supported by Fedora")

        return a

    @classmethod
    def latest_snapshot(cls, config: Config) -> str:
        mirror = config.mirror or "https://kojipkgs.fedoraproject.org"

        url = join_mirror(
            mirror, f"compose/{config.release}/latest-Fedora-{config.release.capitalize()}/COMPOSE_ID"
        )

        return curl(config, url).removeprefix(f"Fedora-{config.release.capitalize()}-").strip()

    @classmethod
    def is_kernel_package(cls, package: str) -> bool:
        return package in ("kernel", "kernel-core")



# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import contextlib
import dataclasses
import logging
import os
import platform
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Optional, cast

import mkosi.resources
from mkosi.config import DocFormat, InitrdProfile, OutputFormat
from mkosi.documentation import show_docs
from mkosi.log import ARG_DEBUG, ARG_DEBUG_SHELL, die, log_notice, log_setup
from mkosi.run import find_binary, run, uncaught_exception_handler
from mkosi.sandbox import __version__, umask
from mkosi.tree import copy_tree, move_tree, rmtree
from mkosi.util import PathString, mandatory_variable, resource_path


@dataclasses.dataclass(frozen=True)
class KernelInstallContext:
    command: str
    kernel_version: str
    entry_dir: Path
    kernel_image: Path
    initrds: list[Path]
    staging_area: Path
    layout: str
    image_type: str
    initrd_generator: Optional[str]
    uki_generator: Optional[str]
    verbose: bool

    @staticmethod
    def parse(*, name: str, description: str) -> "KernelInstallContext":
        parser = argparse.ArgumentParser(
            description=description,
            allow_abbrev=False,
            usage=f"{name} COMMAND KERNEL_VERSION ENTRY_DIR KERNEL_IMAGE…",
        )

        parser.add_argument(
            "command",
            metavar="COMMAND",
            help="The action to perform. Only 'add' is supported.",
        )
        parser.add_argument(
            "kernel_version",
            metavar="KERNEL_VERSION",
            help="Kernel version string",
        )
        parser.add_argument(
            "entry_dir",
            metavar="ENTRY_DIR",
            type=Path,
            nargs="?",
            help="Type#1 entry directory (ignored)",
        )
        parser.add_argument(
            "kernel_image",
            metavar="KERNEL_IMAGE",
            type=Path,
            nargs="?",
            help="Kernel image",
        )
        parser.add_argument(
            "initrds",
            metavar="INITRD…",
            type=Path,
            nargs="*",
            help="Initrd files",
        )
        parser.add_argument(
            "--version",
            action="version",
            version=f"mkosi {__version__}",
        )

        args = parser.parse_args()

        return KernelInstallContext(
            command=args.command,
            kernel_version=args.kernel_version,
            entry_dir=args.entry_dir,
            kernel_image=args.kernel_image,
            initrds=args.initrds,
            staging_area=Path(mandatory_variable("KERNEL_INSTALL_STAGING_AREA")),
            layout=mandatory_variable("KERNEL_INSTALL_LAYOUT"),
            image_type=mandatory_variable("KERNEL_INSTALL_IMAGE_TYPE"),
            initrd_generator=os.getenv("KERNEL_INSTALL_INITRD_GENERATOR"),
            uki_generator=os.getenv("KERNEL_INSTALL_UKI_GENERATOR"),
            verbose=int(os.getenv("KERNEL_INSTALL_VERBOSE", 0)) > 0,
        )


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="mkosi-initrd",
        description="Build initrds or unified kernel images for the current system using mkosi",
        allow_abbrev=False,
        usage="mkosi-initrd [options...]",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="NAME",
        help="Output name",
        default="initrd",
    )
    parser.add_argument(
        "--kernel-image",
        metavar="KERNEL_IMAGE",
        help="Kernel image",
        type=Path,
    )
    parser.add_argument(
        "-t",
        "--format",
        choices=[str(OutputFormat.cpio), str(OutputFormat.uki), str(OutputFormat.directory)],
        help="Output format (CPIO archive, UKI or local directory)",
        default="cpio",
    )
    parser.add_argument(
        "-g",
        "--generic",
        help="Build a generic initrd without host-specific kernel modules",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--profile",
        choices=InitrdProfile.values(),
        help="Which profiles to enable for the initrd",
        action="append",
        default=[],
    )

    initrd_common_args(parser)
    return parser


def is_valid_modulesd(modulesd: Path) -> bool:
    # Check whether a provided kernel modules directory is valid
    return modulesd.is_dir() and (
        (modulesd / "modules.dep").exists() or (modulesd / "modules.dep.bin").exists()
    )


def weak_modules(modulesd: Path) -> list[str]:
    return [
        f"--extra-tree={m.resolve()}:{m.resolve()}"
        for m in (modulesd / "weak-updates").rglob("*.ko*")
        if m.is_symlink()
    ]


def process_crypttab(staging_dir: Path) -> list[str]:
    cmdline = []

    # Generate crypttab with all the x-initrd.attach entries
    if Path("/etc/crypttab").exists():
        try:
            crypttab = [
                line
                for line in Path("/etc/crypttab").read_text().splitlines()
                if (
                    len(entry := line.split()) >= 4
                    and not entry[0].startswith("#")
                    and "x-initrd.attach" in entry[3]
                )
            ]
            if crypttab:
                with (staging_dir / "crypttab").open("w") as f:
                    f.write("# Automatically generated by mkosi-initrd\n")
                    f.write("\n".join(crypttab))
                cmdline += ["--extra-tree", f"{staging_dir / 'crypttab'}:/etc/crypttab"]

                # Add key files
                for line in crypttab:
                    entry = line.split()
                    if (
                        entry[2] in ["-", "none"]
                        and Path(keyfile := f"/etc/cryptsetup-keys.d/{entry[0]}.key").exists()
                    ) or Path(keyfile := entry[2]).exists():
                        cmdline += ["--extra-tree", f"{keyfile}:{keyfile}"]

        except PermissionError:
            logging.warning("Permission denied to access /etc/crypttab, the initrd may be unbootable")

    return cmdline


def network_config() -> list[str]:
    return [
        f"--extra-tree={f}:{f}"
        for f in (
            "/etc/systemd/network",
            "/etc/systemd/networkd.conf",
            "/etc/systemd/networkd.conf.d",
            "/etc/systemd/resolved.conf",
            "/etc/systemd/resolved.conf.d",
        )
        if Path(f).exists()
    ]


def nfs_config() -> list[str]:
    return [f"--extra-tree={f}:{f}" for f in ("/etc/idmapd.conf", "/etc/idmapd.conf.d") if Path(f).exists()]


def raid_config() -> list[str]:
    return [
        f"--extra-tree={f}:{f}"
        for f in ("/etc/mdadm.conf", "/etc/mdadm.conf.d", "/etc/mdadm/mdadm.conf", "/etc/mdadm/mdadm.conf.d")
        if Path(f).exists()
    ]


def vconsole_config() -> list[str]:
    return [
        f"--extra-tree={f}:{f}" for f in ("/etc/default/keyboard", "/etc/vconsole.conf") if Path(f).exists()
    ]


def initrd_finalize(staging_dir: Path, output: str, output_dir: Optional[Path]) -> None:
    if output_dir:
        with umask(~0o700) if os.getuid() == 0 else cast(umask, contextlib.nullcontext()):
            Path(output_dir).mkdir(parents=True, exist_ok=True)
    else:
        output_dir = Path.cwd()

    staging = staging_dir / output
    tmp = output_dir / f"{output}.new"
    final = output_dir / output

    log_notice(f"Copying {staging} to {tmp}")
    # mkosi symlinks the expected output image, so dereference it
    try:
        copy_tree(staging.resolve(), tmp)
    except subprocess.CalledProcessError:
        rmtree(tmp)
        raise

    log_notice(f"Moving {tmp} to {final}")
    rmtree(final)
    move_tree(tmp, final)


def initrd_common_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-k",
        "--kernel-version",
        metavar="KERNEL_VERSION",
        help="Kernel version string",
        default=platform.uname().release,
    )
    parser.add_argument(
        "-O",
        "--output-dir",
        metavar="DIR",
        help="Output directory",
        default=None,
        type=Path,
    )
    parser.add_argument(
        "--debug",
        help="Turn on debugging output",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--debug-shell",
        help="Spawn debug shell if a sandboxed command fails",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--debug-sandbox",
        help="Run mkosi-sandbox with strace",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-D",
        "--show-documentation",
        help="Show the man page",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"mkosi {__version__}",
    )


def include_system_config(name: str) -> list[str]:
    cmdline = []

    for d in ("/usr/lib", "/usr/local/lib", "/run", "/etc"):
        p = Path(d) / name
        if p.exists():
            cmdline += ["--include", os.fspath(p)]

    return cmdline


@uncaught_exception_handler()
def main() -> None:
    log_setup()

    args = create_parser().parse_args()

    if args.show_documentation:
        with resource_path(mkosi.resources) as r:
            show_docs("mkosi-initrd", DocFormat.all(), resources=r)
        return

    modulesd = Path("/usr/lib/modules") / args.kernel_version
    if not is_valid_modulesd(modulesd):
        die(f"Invalid kernel directory: {modulesd}")

    with (
        tempfile.TemporaryDirectory() as staging_dir,
        tempfile.TemporaryDirectory() as sandbox_tree,
    ):
        cmdline: list[PathString] = [
            "mkosi",
            "--force",
            "--directory=",
            f"--format={args.format}",
            f"--output={args.output}",
            f"--output-directory={staging_dir}",
            f"--extra-tree={modulesd}:{modulesd}",
            "--extra-tree=/usr/lib/firmware:/usr/lib/firmware",
            "--remove-files=/usr/lib/firmware/*-ucode",
            "--build-sources=",
            "--include=mkosi-initrd",
        ]  # fmt: skip

        if not args.generic:
            cmdline += ["--kernel-modules=host"]

        cmdline += weak_modules(modulesd)

        for p in args.profile:
            cmdline += ["--profile", p]
            if p == "network":
                cmdline += network_config()
            elif p == "nfs":
                cmdline += nfs_config()
            elif p == "raid":
                cmdline += raid_config()

        if args.kernel_image:
            cmdline += [
                f"--extra-tree={args.kernel_image}:{modulesd}/vmlinuz",
            ]

        if args.debug:
            ARG_DEBUG.set(args.debug)
            cmdline += ["--debug"]
        if args.debug_shell:
            ARG_DEBUG_SHELL.set(args.debug_shell)
            cmdline += ["--debug-shell"]
        if args.debug_sandbox:
            cmdline += ["--debug-sandbox"]

        if os.getuid() == 0:
            cmdline += [
                "--package-cache-dir=/var",
                "--cache-only=metadata",
            ]
            if args.format != OutputFormat.directory.value:
                cmdline += ["--output-mode=600"]

        cmdline += include_system_config("mkosi-initrd")

        # Make sure we don't use any of mkosi's default repositories.
        for p in (
            "yum.repos.d/mkosi.repo",
            "apt/sources.list.d/mkosi.sources",
            "zypp/repos.d/mkosi.repo",
            "pacman.conf",
        ):
            (Path(sandbox_tree) / "etc" / p).parent.mkdir(parents=True, exist_ok=True)
            (Path(sandbox_tree) / "etc" / p).touch()

        # Copy in the host's package manager configuration.
        for p in (
            "dnf",
            "yum.repos.d/",
            "pki/rpm-gpg",
            "apt",
            "zypp",
            "pacman.conf",
            "pacman.d/",
        ):
            if not (Path("/etc") / p).exists():
                continue

            (Path(sandbox_tree) / "etc" / p).parent.mkdir(parents=True, exist_ok=True)
            if (Path("/etc") / p).resolve().is_file():
                shutil.copy2(Path("/etc") / p, Path(sandbox_tree) / "etc" / p)
            else:
                shutil.copytree(
                    Path("/etc") / p,
                    Path(sandbox_tree) / "etc" / p,
                    # If we're running as root, use the keyring from the host, but make sure we don't try to
                    # copy any gpg-agent sockets that might be in /etc/pacman.d/gnupg. If we're not running
                    # as root, we might not have the necessary permissions to access the keyring so don't try
                    # to copy the keyring in that case.
                    ignore=shutil.ignore_patterns("S.*" if os.getuid() == 0 else "gnupg"),
                    dirs_exist_ok=True,
                )

        cmdline += [f"--sandbox-tree={sandbox_tree}"]

        cmdline += process_crypttab(Path(staging_dir))

        if Path("/etc/kernel/cmdline").exists():
            cmdline += ["--kernel-command-line", Path("/etc/kernel/cmdline").read_text()]

        cmdline += vconsole_config()

        # Resolve dnf binary to determine which version the host uses by default
        # (to avoid preferring dnf5 if the host uses dnf4)
        # as there's a much bigger chance that it has a populated dnf cache directory.
        run(
            cmdline,
            stdin=sys.stdin,
            stdout=sys.stdout,
            env=os.environ | ({"MKOSI_DNF": dnf.resolve().name} if (dnf := find_binary("dnf")) else {}),
        )

        initrd_finalize(Path(staging_dir), args.output, args.output_dir)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

import dataclasses
import shutil
import textwrap
from collections.abc import Sequence
from contextlib import AbstractContextManager
from pathlib import Path

from mkosi.config import Config
from mkosi.context import Context
from mkosi.distribution import detect_distribution
from mkosi.installer import PackageManager
from mkosi.log import complete_step
from mkosi.run import CompletedProcess, run, workdir
from mkosi.sandbox import umask
from mkosi.tree import copy_tree
from mkosi.util import _FILE, PathString
from mkosi.versioncomp import GenericVersion


@dataclasses.dataclass(frozen=True)
class PacmanRepository:
    id: str
    url: str


class Pacman(PackageManager):
    @classmethod
    def executable(cls, config: Config) -> str:
        return "pacman"

    @classmethod
    def subdir(cls, config: Config) -> Path:
        return Path("pacman")

    @classmethod
    def package_subdirs(cls, cache: Path) -> list[tuple[Path, Path]]:
        return [(Path("pkg"), Path("pkg"))]

    @classmethod
    def package_globs(cls) -> list[str]:
        return ["*.pkg.tar*"]

    @classmethod
    def state_subdirs(cls) -> list[Path]:
        return [Path("sync")]

    @classmethod
    def scripts(cls, context: Context) -> dict[str, list[PathString]]:
        return {
            "pacman": cls.apivfs_script_cmd(context) + cls.env_cmd(context) + cls.cmd(context),
            "mkosi-install":   ["pacman", "--sync", "--needed"],
            "mkosi-upgrade":   ["pacman", "--sync", "--sysupgrade", "--needed"],
            "mkosi-remove":    ["pacman", "--remove", "--recursive", "--nosave"],
            "mkosi-reinstall": ["pacman", "--sync"],
        }  # fmt: skip

    @classmethod
    def mounts(cls, context: Context) -> list[PathString]:
        mounts = [
            *super().mounts(context),
            # pacman writes downloaded packages to the first writable cache directory. We don't want it to
            # write to our local repository directory so we expose it as a read-only directory to pacman.
            "--ro-bind", context.repository, "/var/cache/pacman/mkosi",
        ]  # fmt: skip

        if any(context.keyring_dir.iterdir()):
            mounts += ["--ro-bind", context.keyring_dir, "/etc/pacman.d/gnupg"]

        if (context.root / "var/lib/pacman/local").exists():
            # pacman reuses the same directory for the sync databases and the local database containing the
            # list of installed packages. The former should go in the cache directory, the latter should go
            # in the image, so we bind mount the local directory from the image to make sure that happens. We
            # make sure to bind mount directly from the mounted /buildroot directly instead of from the host
            # root directory since /buildroot might be an overlay mount and we want to make sure any writes
            # are done to the upperdir of the overlay mount.
            mounts += ["--bind", "+/buildroot/var/lib/pacman/local", "/var/lib/pacman/local"]

        return mounts

    @classmethod
    def setup(cls, context: Context, repositories: Sequence[PacmanRepository]) -> None:
        if context.config.repository_key_check:
            sig_level = "Required DatabaseOptional"
        else:
            # If we are using a single local mirror built on the fly there
            # will be no signatures
            sig_level = "Never"

        (context.sandbox_tree / "etc/mkosi-local.conf").touch()

        config = context.sandbox_tree / "etc/pacman.conf"
        if config.exists():
            # If DownloadUser is specified, remove it as the user won't be available in the sandbox.
            lines = config.read_text().splitlines()
            lines = [line for line in lines if not line.strip().startswith("DownloadUser")]
            config.write_text("\n".join(lines))
            return

        config.parent.mkdir(exist_ok=True, parents=True)

        with config.open("w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    [options]
                    SigLevel = {sig_level}
                    LocalFileSigLevel = Optional
                    ParallelDownloads = 5
                    Architecture = {cls.architecture(context)}

                    """
                )
            )

            if not context.config.with_docs:
                f.write(
                    textwrap.dedent(
                        """\
                        NoExtract = usr/share/doc/*
                        NoExtract = usr/share/man/*
                        NoExtract = usr/share/groff/*
                        NoExtract = usr/share/gtk-doc/*
                        NoExtract = usr/share/info/*
                        """
                    )
                )

            # This has to go first so that our local repository always takes precedence over any other ones.
            f.write("Include = /etc/mkosi-local.conf\n")

            if any((context.sandbox_tree / "etc/pacman.d/").glob("*.conf")):
                f.write(
                    textwrap.dedent(
                        """\

                        Include = /etc/pacman.d/*.conf
                        """
                    )
                )

            for repo in repositories:
                f.write(
                    textwrap.dedent(
                        f"""\

                        [{repo.id}]
                        Server = {repo.url}
                        """
                    )
                )

    @classmethod
    def cmd(cls, context: Context) -> list[PathString]:
        return [
            "pacman",
            "--root=/buildroot",
            "--logfile=/dev/null",
            "--dbpath=/var/lib/pacman",
            # Make sure pacman looks at our local repository first by putting it as the first cache
            # directory. We mount it read-only so the second directory will still be used for writing new
            # cache entries.
            "--cachedir=/var/cache/pacman/mkosi",
            "--cachedir=/var/cache/pacman/pkg",
            "--hookdir=/buildroot/etc/pacman.d/hooks",
            "--arch", cls.architecture(context),
            "--color", "auto",
            "--noconfirm",
        ]  # fmt: skip

    @classmethod
    def invoke(
        cls,
        context: Context,
        operation: str,
        arguments: Sequence[str] = (),
        *,
        apivfs: bool = False,
        stdout: _FILE = None,
    ) -> CompletedProcess:
        with umask(~0o755):
            (context.root / "var/lib/pacman/local").mkdir(parents=True, exist_ok=True)
            (context.root / "etc/pacman.d/hooks").mkdir(parents=True, exist_ok=True)
            (
                context.config.package_cache_dir_or_default() / "cache" / cls.subdir(context.config) / "pkg"
            ).mkdir(parents=True, exist_ok=True)

        return run(
            cls.cmd(context) + [operation, *arguments],
            sandbox=cls.sandbox(context, apivfs=apivfs),
            env=cls.finalize_environment(context),
            stdout=stdout,
        )

    @classmethod
    def install(
        cls,
        context: Context,
        packages: Sequence[str],
        *,
        apivfs: bool = True,
        allow_downgrade: bool = False,
    ) -> None:
        arguments = ["--needed", "--assume-installed", "initramfs"]

        if allow_downgrade:
            arguments += ["--sysupgrade", "--sysupgrade"]

        arguments += [*packages]

        cls.invoke(context, "--sync", arguments, apivfs=apivfs)

    @classmethod
    def remove(cls, context: Context, packages: Sequence[str]) -> None:
        installed = {
            cls.parse_pkg_desc(i)[0] for i in (context.root / "var/lib/pacman/local").glob("*/desc")
        }
        remove = [p for p in packages if p in installed]
        if remove:
            cls.invoke(context, "--remove", ["--nosave", "--recursive", *remove], apivfs=True)

    @classmethod
    def keyring(cls, context: Context) -> None:
        def sandbox() -> AbstractContextManager[list[PathString]]:
            return cls.sandbox(
                context,
                apivfs=False,
                # By default the keyring is mounted read-only so we override the read-only mount with a
                # writable mount to make it writable for the following pacman-key commands.
                options=["--bind", context.keyring_dir, "/etc/pacman.d/gnupg"],
            )

        if (
            (d := detect_distribution(context.config.tools())[0])
            and d.is_apt_distribution()
            and (context.sandbox_tree / "usr/share/pacman/keyrings").exists()
        ):
            # pacman on Debian/Ubuntu looks for keyrings in /usr/share/keyrings so make sure all sandbox
            # trees keyrings are available in that location as well.
            (context.sandbox_tree / "usr/share").mkdir(parents=True, exist_ok=True)
            copy_tree(
                context.sandbox_tree / "usr/share/pacman/keyrings",
                context.sandbox_tree / "usr/share/keyrings",
                dereference=True,
                sandbox=context.sandbox,
            )

        context.keyring_dir.mkdir(parents=True, exist_ok=True)

        with complete_step("Populating pacman keyring"):
            run(["pacman-key", "--init"], sandbox=sandbox())
            run(["pacman-key", "--populate"], sandbox=sandbox())

    @classmethod
    def sync(cls, context: Context, force: bool) -> None:
        cls.invoke(context, "--sync", ["--refresh", *(["--refresh"] if force else [])])

    @classmethod
    def createrepo(cls, context: Context) -> None:
        run(
            [
                "repo-add",
                "--quiet",
                workdir(context.repository / "mkosi.db.tar"),
                *sorted(
                    (workdir(p) for p in context.repository.glob("*.pkg.tar*")),
                    key=lambda p: GenericVersion(Path(p).name),
                ),
            ],
            sandbox=context.sandbox(options=["--bind", context.repository, workdir(context.repository)]),
        )

        (context.sandbox_tree / "etc/mkosi-local.conf").write_text(
            textwrap.dedent(
                """\
                [mkosi]
                Server = file:///i/dont/exist
                SigLevel = Never
                Usage = Install Search Upgrade
                """
            )
        )

        # pacman can't sync a single repository, so we go behind its back and do it ourselves.
        shutil.move(context.repository / "mkosi.db.tar", context.metadata_dir / "lib/pacman/sync/mkosi.db")

    @classmethod
    def parse_pkg_desc(cls, path: Path) -> tuple[str, str, str, str]:
        name = version = base = arch = ""

        with path.open() as desc:
            for line in desc:
                line = line.strip()
                if line == "%NAME%":
                    name = next(desc).strip()
                elif line == "%VERSION%":
                    version = next(desc).strip()
                elif line == "%BASE%":
                    base = next(desc).strip()
                elif line == "%ARCH%":
                    arch = next(desc).strip()
                    break

        return name, version, base, arch



# SPDX-License-Identifier: LGPL-2.1-or-later

from mkosi.context import Context
from mkosi.distribution import Distribution, centos, join_mirror
from mkosi.installer.rpm import RpmRepository, find_rpm_gpgkey
from mkosi.log import die


class Installer(centos.Installer, distribution=Distribution.rocky):
    @classmethod
    def pretty_name(cls) -> str:
        return "Rocky Linux"

    @classmethod
    def gpgurls(cls, context: Context) -> tuple[str, ...]:
        major = cls.major_release(context.config)
        return (
            find_rpm_gpgkey(
                context,
                f"RPM-GPG-KEY-Rocky-{major}",
                f"https://download.rockylinux.org/pub/rocky/RPM-GPG-KEY-Rocky-{major}",
            ),
        )

    @classmethod
    def repository_variants(
        cls,
        context: Context,
        gpgurls: tuple[str, ...],
        repo: str,
    ) -> list[RpmRepository]:
        if context.config.snapshot:
            die(f"Snapshot= is not supported for {cls.pretty_name()}")

        if context.config.mirror:
            url = f"baseurl={join_mirror(context.config.mirror, f'$releasever/{repo}/$basearch/os')}"
        else:
            url = f"mirrorlist=https://mirrors.rockylinux.org/mirrorlist?arch=$basearch&repo={repo}-$releasever"

        return [RpmRepository(repo, url, gpgurls)]

    @classmethod
    def sig_repositories(cls, context: Context) -> list[RpmRepository]:
        return []



# SPDX-License-Identifier: LGPL-2.1-or-later
# PYTHON_ARGCOMPLETE_OK

import faulthandler
import signal
import sys
from types import FrameType
from typing import Optional

import mkosi.resources
from mkosi import run_verb
from mkosi.config import parse_config
from mkosi.log import log_setup, stash_terminal_title
from mkosi.run import find_binary, run, uncaught_exception_handler
from mkosi.util import resource_path

INTERRUPTED = False


def onsignal(signal: int, frame: Optional[FrameType]) -> None:
    global INTERRUPTED
    if INTERRUPTED:
        return

    INTERRUPTED = True
    raise KeyboardInterrupt()


@uncaught_exception_handler()
def main() -> None:
    signal.signal(signal.SIGINT, onsignal)
    signal.signal(signal.SIGTERM, onsignal)
    signal.signal(signal.SIGHUP, onsignal)

    log_setup()

    with resource_path(mkosi.resources) as resources, stash_terminal_title():
        args, tools, images = parse_config(sys.argv[1:], resources=resources)

        if args.debug:
            faulthandler.enable()

        try:
            run_verb(args, tools, images, resources=resources)
        finally:
            if sys.stderr.isatty() and find_binary("tput"):
                run(["tput", "cnorm"], check=False)
                run(["tput", "smam"], check=False)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

import dataclasses
import textwrap
from pathlib import Path
from typing import Literal, Optional, overload

from mkosi.context import Context
from mkosi.distribution import Distribution
from mkosi.log import die
from mkosi.run import glob_in_sandbox
from mkosi.util import PathString


@dataclasses.dataclass(frozen=True)
class RpmRepository:
    id: str
    url: str
    gpgurls: tuple[str, ...]
    enabled: bool = True
    sslcacert: Optional[Path] = None
    sslclientkey: Optional[Path] = None
    sslclientcert: Optional[Path] = None
    priority: Optional[int] = None


@overload
def find_rpm_gpgkey(
    context: Context,
    key: str,
    fallback: Optional[str] = None,
    *,
    required: Literal[True] = True,
) -> str: ...


@overload
def find_rpm_gpgkey(
    context: Context,
    key: str,
    fallback: Optional[str] = None,
    *,
    required: bool,
) -> Optional[str]: ...


def find_rpm_gpgkey(
    context: Context,
    key: str,
    fallback: Optional[str] = None,
    *,
    required: bool = True,
) -> Optional[str]:
    # We assume here that GPG keys will only ever be relative symlinks and never absolute symlinks.

    paths = glob_in_sandbox(
        f"/usr/share/distribution-gpg-keys/*/{key}*",
        f"/etc/pki/rpm-gpg/{key}*",
        sandbox=context.sandbox(),
    )

    if paths:
        return Path(paths[0]).as_uri()

    if fallback and context.config.repository_key_fetch:
        return fallback

    if required:
        die(
            f"{key} GPG key not found in /usr/share/distribution-gpg-keys or /etc/pki/rpm-gpg",
            hint="Make sure the distribution-gpg-keys package is installed",
        )

    return None


def setup_rpm(
    context: Context,
    *,
    dbpath: str = "/usr/lib/sysimage/rpm",
    dbbackend: Optional[str] = None,
) -> None:
    confdir = context.sandbox_tree / "etc/rpm"
    confdir.mkdir(parents=True, exist_ok=True)
    if not (confdir / "macros.lang").exists() and context.config.locale:
        (confdir / "macros.lang").write_text(f"%_install_langs {context.config.locale}")

    if not (confdir / "macros.dbpath").exists():
        (confdir / "macros.dbpath").write_text(f"%_dbpath {dbpath}")

    if dbbackend:
        (confdir / "macros.db_backend").write_text(f"%_db_backend {dbbackend}")

    if context.config.distribution == Distribution.opensuse or (
        context.config.distribution.is_centos_variant() and context.config.release == "9"
    ):
        # Write an rpm sequoia policy that makes sure "sha1.second_preimage_resistance = always" is
        # configured and makes sure that a minimal config is in place to make sure builds succeed.
        # TODO: Remove when distributions GPG keys are accepted by the default rpm-sequoia config everywhere.

        p = context.sandbox_tree / "etc/crypto-policies/back-ends/rpm-sequoia.config"
        p.parent.mkdir(parents=True, exist_ok=True)
        prev = p.read_text() if p.exists() else ""

        with p.open("w") as f:
            for line in prev.splitlines(keepends=True):
                if line.startswith("sha1.second_preimage_resistance"):
                    f.write('sha1.second_preimage_resistance = "always"\n')
                else:
                    f.write(line)

            if not any(line.startswith("[hash_algorithms]") for line in prev.splitlines()):
                f.write(
                    textwrap.dedent(
                        """
                        [hash_algorithms]
                        sha1.second_preimage_resistance = "always"
                        sha224 = "always"
                        sha256 = "always"
                        sha384 = "always"
                        sha512 = "always"
                        default_disposition = "never"
                        """
                    )
                )


def rpm_cmd() -> list[PathString]:
    return ["env", "HOME=/", "rpm", "--root=/buildroot"]



# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import errno
import logging
import os
import shutil
import subprocess
import tempfile
from collections.abc import Iterator
from pathlib import Path

from mkosi.config import ConfigFeature
from mkosi.log import die
from mkosi.run import SandboxProtocol, nosandbox, run, workdir
from mkosi.sandbox import (
    BTRFS_SUPER_MAGIC,
    FS_NOCOW_FL,
    OVERLAYFS_SUPER_MAGIC,
    btrfs_subvol_create,
    btrfs_subvol_delete,
    btrfs_subvol_snapshot,
    chattr,
    lsattr,
    statfs,
)
from mkosi.util import PathString, flatten
from mkosi.versioncomp import GenericVersion


def is_subvolume(path: Path) -> bool:
    return path.is_dir() and path.stat().st_ino == 256 and statfs(os.fspath(path)) == BTRFS_SUPER_MAGIC


def cp_version(*, sandbox: SandboxProtocol = nosandbox) -> GenericVersion:
    return GenericVersion(
        run(
            ["cp", "--version"],
            sandbox=sandbox(),
            stdout=subprocess.PIPE,
        )
        .stdout.splitlines()[0]
        .split()[3]
    )


def make_tree(path: Path, *, use_subvolumes: ConfigFeature = ConfigFeature.disabled) -> Path:
    path = path.absolute()

    if statfs(os.fspath(path.parent)) != BTRFS_SUPER_MAGIC:
        if use_subvolumes == ConfigFeature.enabled:
            die(f"Subvolumes requested but {path} is not located on a btrfs filesystem")

        path.mkdir()
        return path

    if use_subvolumes != ConfigFeature.disabled:
        try:
            btrfs_subvol_create(os.fspath(path))
            return path
        except OSError as e:
            if use_subvolumes == ConfigFeature.enabled:
                raise e

            if e.errno != errno.ENOTTY:
                logging.debug(f"Failed to create subvolume {path}, using a regular directory: {e}")

    path.mkdir()
    return path


@contextlib.contextmanager
def preserve_target_directories_stat(src: Path, dst: Path) -> Iterator[None]:
    dirs = [p for d in src.glob("**/") if (dst / (p := d.relative_to(src))).exists()]

    with tempfile.TemporaryDirectory() as tmp:
        for d in dirs:
            (tmp / d).mkdir(exist_ok=True)
            shutil.copystat(dst / d, tmp / d)

        yield

        for d in dirs:
            shutil.copystat(tmp / d, dst / d)


def maybe_make_nocow(path: Path) -> None:
    try:
        chattr(os.fspath(path), lsattr(os.fspath(path)) | FS_NOCOW_FL)
    except OSError as e:
        if e.errno not in (errno.ENOTTY, errno.EOPNOTSUPP, errno.EINVAL):
            raise


def tree_has_selinux_xattr(path: Path) -> bool:
    return any(
        "security.selinux" in os.listxattr(p, follow_symlinks=False) for p in (path, *path.rglob("*"))
    )


def copy_tree(
    src: Path,
    dst: Path,
    *,
    preserve: bool = True,
    dereference: bool = False,
    use_subvolumes: ConfigFeature = ConfigFeature.disabled,
    sandbox: SandboxProtocol = nosandbox,
) -> Path:
    src = src.absolute()
    dst = dst.absolute()

    options: list[PathString] = [
        "--ro-bind", src, workdir(src, sandbox),
        "--bind", dst.parent, workdir(dst.parent, sandbox),
    ]  # fmt: skip

    attrs = "mode,links"
    if preserve:
        attrs += ",timestamps,ownership"

        # Trying to copy selinux xattrs to overlayfs fails with "Operation not supported" in containers.
        if statfs(os.fspath(dst.parent)) != OVERLAYFS_SUPER_MAGIC or not tree_has_selinux_xattr(src):
            attrs += ",xattr"

    def copy() -> None:
        if src.is_file():
            try:
                attr = lsattr(os.fspath(src))
            except OSError:
                attr = 0

            if attr & FS_NOCOW_FL:
                fdst = dst / src.name if dst.is_dir() else dst
                fdst.touch()
                maybe_make_nocow(fdst)

        cmdline: list[PathString] = [
            "cp",
            "--recursive",
            "--dereference" if dereference else "--no-dereference",
            f"--preserve={attrs}",
            "--reflink=auto",
            "--copy-contents",
            workdir(src, sandbox),
            workdir(dst, sandbox),
        ]

        if dst.exists() and dst.is_dir() and any(dst.iterdir()) and cp_version(sandbox=sandbox) >= "9.5":
            cmdline += ["--keep-directory-symlink"]

        # If the source and destination are both directories, we want to merge the source directory with the
        # destination directory. If the source if a file and the destination is a directory, we want to copy
        # the source inside the directory.
        if src.is_dir():
            cmdline += ["--no-target-directory"]

        run(cmdline, sandbox=sandbox(options=options))

    # Subvolumes always have inode 256 so we can use that to check if a directory is a subvolume.
    if (
        use_subvolumes == ConfigFeature.disabled
        or not preserve
        or not is_subvolume(src)
        or statfs(os.fspath(dst.parent)) != BTRFS_SUPER_MAGIC
        or (dst.exists() and (not dst.is_dir() or any(dst.iterdir())))
    ):
        with preserve_target_directories_stat(src, dst) if not preserve else contextlib.nullcontext():
            copy()

        return dst

    # btrfs can't snapshot to an existing directory so make sure the destination does not exist.
    if dst.exists():
        dst.rmdir()

    try:
        btrfs_subvol_snapshot(os.fspath(src), os.fspath(dst))
    except OSError as e:
        if use_subvolumes == ConfigFeature.enabled:
            raise e

        if e.errno != errno.ENOTTY:
            logging.debug(f"Failed to snapshot subvolume {src} to {dst}, falling back to copying: {e}")

        with preserve_target_directories_stat(src, dst) if not preserve else contextlib.nullcontext():
            copy()

    return dst


def rmtree(*paths: Path, sandbox: SandboxProtocol = nosandbox) -> None:
    if not paths:
        return

    paths = tuple(p.absolute() for p in paths)

    # Silence and ignore failures since when not running as root, this will fail with a permission error
    # unless the btrfs filesystem is mounted with user_subvol_rm_allowed.
    for p in sorted({p for p in paths if not p.is_symlink() and p.exists() and is_subvolume(p)}):
        try:
            # Make sure the subvolume is writable which is required to be able to remove it.
            os.chmod(p, 755)
            btrfs_subvol_delete(os.fspath(p))
        except OSError as e:
            if e.errno in (errno.EPERM, errno.EACCES):
                logging.debug(
                    f"Could not delete subvolume {p} due to permission issues.\n"
                    "Consider adding user_subvol_rm_allowed to your btrfs filesystem mount options."
                )
            else:
                logging.debug(f"Failed to delete subvolume {p}: {e}")

    filtered = sorted({p for p in paths if p.exists() or p.is_symlink()})
    if filtered:
        run(
            ["rm", "-rf", "--", *(workdir(p, sandbox) for p in filtered)],
            sandbox=sandbox(
                options=flatten(("--bind", p.parent, workdir(p.parent, sandbox)) for p in filtered),
            ),
        )


def move_tree(
    src: Path,
    dst: Path,
    *,
    use_subvolumes: ConfigFeature = ConfigFeature.disabled,
    sandbox: SandboxProtocol = nosandbox,
) -> Path:
    src = src.absolute()
    dst = dst.absolute()

    if src == dst:
        return dst

    if dst.is_dir():
        dst = dst / src.name

    try:
        src.rename(dst)
    except OSError as e:
        if e.errno != errno.EXDEV:
            raise e

        logging.info(
            f"Could not rename {src} to {dst} as they are located on different devices, "
            "falling back to copying"
        )
        copy_tree(src, dst, use_subvolumes=use_subvolumes, sandbox=sandbox)
        rmtree(src, sandbox=sandbox)

    return dst



# SPDX-License-Identifier: LGPL-2.1-or-later

import textwrap
from collections.abc import Sequence
from pathlib import Path

from mkosi.config import Config, yes_no
from mkosi.context import Context
from mkosi.installer import PackageManager
from mkosi.installer.rpm import RpmRepository, rpm_cmd
from mkosi.log import ARG_DEBUG
from mkosi.run import CompletedProcess, run, workdir
from mkosi.util import _FILE, PathString


class Zypper(PackageManager):
    @classmethod
    def executable(cls, config: Config) -> str:
        return "zypper"

    @classmethod
    def subdir(cls, config: Config) -> Path:
        return Path("zypp")

    @classmethod
    def package_subdirs(cls, cache: Path) -> list[tuple[Path, Path]]:
        return [(Path("packages"), Path("packages"))]

    @classmethod
    def package_globs(cls) -> list[str]:
        return ["*.rpm"]

    @classmethod
    def scripts(cls, context: Context) -> dict[str, list[PathString]]:
        install: list[PathString] = [
            "zypper",
            "install",
            "--download", "in-advance",
            "--recommends" if context.config.with_recommends else "--no-recommends",
            "--force-resolution",
        ]  # fmt: skip

        return {
            "zypper": cls.apivfs_script_cmd(context) + cls.env_cmd(context) + cls.cmd(context),
            "rpm":    cls.apivfs_script_cmd(context) + rpm_cmd(),
            "mkosi-install":   install,
            "mkosi-upgrade":   ["zypper", "update"],
            "mkosi-remove":    ["zypper", "--ignore-unknown", "remove", "--clean-deps"],
            "mkosi-reinstall": install + ["--force"],
        }  # fmt: skip

    @classmethod
    def setup(cls, context: Context, repositories: Sequence[RpmRepository]) -> None:
        config = context.sandbox_tree / "etc/zypp/zypp.conf"
        config.parent.mkdir(exist_ok=True, parents=True)

        # rpm.install.excludedocs can only be configured in zypp.conf so we append to any user provided
        # config file. Let's also bump the refresh delay to the same default as dnf which is 48 hours.
        with config.open("a") as f:
            f.write(
                textwrap.dedent(
                    f"""
                    [main]
                    rpm.install.excludedocs = {yes_no(not context.config.with_docs)}
                    repo.refresh.delay = {48 * 60}
                    """
                )
            )

        repofile = context.sandbox_tree / "etc/zypp/repos.d/mkosi.repo"
        if not repofile.exists():
            repofile.parent.mkdir(exist_ok=True, parents=True)
            with repofile.open("w") as f:
                for repo in repositories:
                    f.write(
                        textwrap.dedent(
                            f"""\
                            [{repo.id}]
                            name={repo.id}
                            {repo.url}
                            gpgcheck=1
                            enabled={int(repo.enabled)}
                            autorefresh=0
                            keeppackages=1
                            """
                        )
                    )

                    if repo.priority:
                        f.write(f"priority={repo.priority}\n")

                    for i, url in enumerate(repo.gpgurls):
                        f.write("gpgkey=" if i == 0 else len("gpgkey=") * " ")
                        f.write(f"{url}\n")

                    f.write("\n")

    @classmethod
    def finalize_environment(cls, context: Context) -> dict[str, str]:
        return super().finalize_environment(context) | {
            "ZYPP_CONF": "/etc/zypp/zypp.conf",
            "RPM_FORCE_DEBIAN": "1",
        }

    @classmethod
    def cmd(cls, context: Context) -> list[PathString]:
        return [
            "zypper",
            "--installroot=/buildroot",
            "--cache-dir=/var/cache/zypp",
            "--non-interactive",
            "--no-refresh",
            f"--releasever={context.config.release}",
            *(["--gpg-auto-import-keys"] if context.config.repository_key_fetch else []),
            *(["--no-gpg-checks"] if not context.config.repository_key_check else []),
            *([f"--plus-content={repo}" for repo in context.config.repositories]),
            *(["-vv"] if ARG_DEBUG.get() else []),
        ]

    @classmethod
    def invoke(
        cls,
        context: Context,
        operation: str,
        arguments: Sequence[str] = (),
        *,
        options: Sequence[str] = (),
        apivfs: bool = False,
        stdout: _FILE = None,
    ) -> CompletedProcess:
        return run(
            cls.cmd(context) + [*options, operation, *arguments],
            sandbox=cls.sandbox(context, apivfs=apivfs),
            env=cls.finalize_environment(context),
            stdout=stdout,
        )

    @classmethod
    def install(
        cls,
        context: Context,
        packages: Sequence[str],
        *,
        apivfs: bool = True,
        allow_downgrade: bool = False,
    ) -> None:
        arguments = [
            "--download", "in-advance",
            "--recommends" if context.config.with_recommends else "--no-recommends",
            "--force-resolution",
        ]  # fmt: skip

        if allow_downgrade:
            arguments += ["--allow-downgrade"]

        arguments += [*packages]

        cls.invoke(context, "install", arguments, apivfs=apivfs)

    @classmethod
    def remove(cls, context: Context, packages: Sequence[str]) -> None:
        cls.invoke(context, "remove", ["--clean-deps", *packages], apivfs=True, options=["--ignore-unknown"])

    @classmethod
    def sync(cls, context: Context, force: bool, arguments: Sequence[str] = ()) -> None:
        cls.invoke(context, "refresh", [*(["--force"] if force else []), *arguments])

    @classmethod
    def createrepo(cls, context: Context) -> None:
        run(
            ["createrepo_c", workdir(context.repository)],
            sandbox=context.sandbox(options=["--bind", context.repository, workdir(context.repository)]),
        )

        (context.sandbox_tree / "etc/zypp/repos.d/mkosi-local.repo").write_text(
            textwrap.dedent(
                """\
                [mkosi]
                name=mkosi
                baseurl=file:///repository
                gpgcheck=0
                autorefresh=0
                keeppackages=0
                priority=10
                """
            )
        )

        cls.sync(context, force=True, arguments=["mkosi"])
